//
//  ViewController.swift
//  DrawingView
//
//  Created by Scott Riccardelli on 6/22/17.
//  Copyright © 2017 Scott Riccardelli. All rights reserved.
//

import Cocoa

class ViewController: NSViewController {

    // MARK: - OUTLETS
    
    // TOOLS
    @IBOutlet weak var mToolMatrix: NSMatrix!
    
    // matrix of buttons for selecting the drawing tool
    @IBOutlet weak var mToolStickyCheckbox: NSButton!
    
    // checkbox for setting "sticky" state of tools
    
    // STYLE
    
    @IBOutlet weak var mStyleFillCheckbox: NSButton!
    
    // checkbox for enabling the "fill" property
    
    @IBOutlet weak var mStyleFillColourWell: NSColorWell!
    
    // colour well for the fill's colour
    
    @IBOutlet weak var mStyleStrokeCheckbox: NSButton!
    
    // checkbox for enabling the "stroke" property
    
    @IBOutlet weak var mStyleStrokeColourWell: NSColorWell!
    
    // colour well for the stroke's colour
    
    @IBOutlet weak var mStyleStrokeWidthTextField: NSTextField!
    
    // text field for the stroke's width
    
    @IBOutlet weak var mStyleStrokeWidthStepper: NSStepper!
    
    // stepper buttons for the stroke's width
    
    // GRID
    
    @IBOutlet weak var mGridMatrix: NSMatrix!
    
    // matrix of buttons for selecting the grid to use
    
    @IBOutlet weak var mGridSnapCheckbox: NSButton!
    
    // checkbox to enable "snap to grid"
    
    // LAYERS
    
    @IBOutlet weak var mLayerTable: NSTableView!
    
    // table view for listing the drawing's layers
    
    @IBOutlet weak var mLayerAddButton: NSButton!
    
    // button for adding a new layer
    
    @IBOutlet weak var mLayerRemoveButton: NSButton!
    
    // button for removing the active (selected) layer
    
    // MAIN VIEW
    
    // outlet to the main DKDrawingView that displays the content (and owns the drawing)
    
    @IBOutlet weak var mDrawingView: DKDrawingView!
    
    // MARK: -
    
    override func viewDidLoad() {
        super.viewDidLoad()

        // Do any additional setup after loading the view.
    }

    override var representedObject: Any? {
        didSet {
        // Update the view, if already loaded.
        }
    }

    // MARK: - Actions
    
    @IBAction func toolMatrixAction(_ sender: NSMatrix) {
        // the drawing view can handle this for us, provided we pass it an object that responds to -title and returns
        // the valid name of a registered tool. The selected button cell is just such an object.
        let cell: NSButtonCell? = sender.selectedCell() as! NSButtonCell
        mDrawingView.selectDrawingTool(byName: cell)
    }
    
    @IBAction func toolStickyAction(_ sender: Any) {
        // sets the tool controller's flag to the inverted state of the checkbox
        (mDrawingView.controller as? DKToolController)?.automaticallyRevertsToSelectionTool = !CInt(sender)
    }
    
    // MARK: -
    
    @IBAction func styleFillColourAction(_ sender: NSColorWell) {
        // get the style of the selected object
        let style: DKStyle? = styleOfSelectedObject()
        style?.fillColour = sender.color
        mDrawingView.undoManager?.setActionName("Change Fill Colour")
    }
    
    @IBAction func styleStrokeColourAction(_ sender: NSColorWell) {
    
        // get the style of the selected object
        let style: DKStyle? = styleOfSelectedObject()
        style?.strokeColour = sender.color
        mDrawingView.undoManager?.setActionName("Change Stroke Colour")
    }
    
    @IBAction func styleStrokeWidthAction(_ sender: Any) {
        // get the style of the selected object
        let style: DKStyle? = styleOfSelectedObject()
        style?.strokeWidth = CFloat(sender)
        // synchronise the text field and the stepper so they both have the same value
        if sender == mStyleStrokeWidthStepper {
            mStyleStrokeWidthTextField = CFloat(sender)
        }
        else {
            mStyleStrokeWidthStepper = CFloat(sender)
        }
        mDrawingView.undoManager?.setActionName("Change Stroke Width")
    }
    
    @IBAction func styleFillCheckboxAction(_ sender: NSButton) {
        // get the style of the selected object
        let style: DKStyle? = styleOfSelectedObject()
        let removing: Bool = (CInt(sender) == 0)
        if removing {
            style?.fillColour = nil
            mDrawingView.undoManager?.setActionName("Delete Fill")
        }
        else {
            style?.fillColour = mStyleFillColourWell.color
            mDrawingView.undoManager?.setActionName("Add Fill")
        }
    }
    
    @IBAction func styleStrokeCheckboxAction(_ sender: NSButton) {
        // get the style of the selected object
        let style: DKStyle? = styleOfSelectedObject()
        let removing: Bool = (CInt(sender) == 0)
        if removing {
            style?.strokeColour = nil
            mDrawingView.undoManager?.setActionName("Delete Stroke")
        }
        else {
            style?.strokeColour = mStyleStrokeColourWell.color
            mDrawingView.undoManager?.setActionName("Add Stroke")
        }
    }
    
    // MARK: -
    
    @IBAction func gridMatrixAction(_ sender: NSMatrix) {
        // the drawing's grid layer already knows how to do this - just pass it the selected cell from where it
        // can extract the tag which it interprets as one of the standard grids.
        mDrawingView.drawing().gridLayer().measurementSystemAction = sender.selectedCell()
    }
    
    @IBAction func snaptoGridAction(_ sender: NSButton) {
        // set the drawing's snapToGrid flag to match the sender's state
        mDrawingView.drawing().snapsToGrid = CInt(sender)
    }
    
    // MARK: -
    
    @IBAction func layerAddButtonAction(_ sender: NSButton) {
        // adding a new layer - first create it
        let newLayer = DKObjectDrawingLayer()
        // add it to the drawing and make it active - this triggers notifications which update the UI
        mDrawingView.drawing().add(newLayer, andActivateIt: true)
        // drawing now owns the layer so we can release it
        // inform the Undo Manager what we just did:
        mDrawingView.undoManager?.setActionName("New Drawing Layer")
    }
    
    @IBAction func layerRemoveButtonAction(_ sender: NSButton) {
        // removing the active (selected) layer - first find that layer
        let activeLayer: DKLayer? = mDrawingView.drawing().activeLayer()
        // remove it and activate another (passing nil tells the drawing to use its nous to activate something sensible)
        mDrawingView.drawing().remove(activeLayer, andActivateLayer: nil)
        // inform the Undo Manager what we just did:
        mDrawingView.undoManager?.setActionName("Delete Drawing Layer")
    }
    
    // MARK: -
    
    func drawingSelectionDidChange(_ note: Notification) {
        // the selection changed within the drawing - update the UI to match the state of whatever was selected. We pass nil
        // because in fact we just grab the current selection directly.
        updateControls(forSelection: nil)
    }
    
    func activeLayerDidChange(_ note: Notification) {
        // change the selection in the layer table to match the actual layer that has been activated
        let dwg: DKDrawing? = mDrawingView.drawing()
        if dwg != nil {
            // now find the active layer's index and set the selection to the same value
            let index: UInt? = dwg?.index(of: dwg?.activeLayer())
            if index != NSNotFound {
                mLayerTable.selectRowIndexes(IndexSet(index), byExtendingSelection: false)
            }
        }
    }
    
    func numberOfLayersChanged(_ note: Notification) {
        // update the table to match the number of layers in the drawing
        mLayerTable.reloadData()
        // re-establish the correct selection - requires a small delay so that the table is fully reloaded before the
        // selection is changed to avoid a potential out of range exception.
        perform(#selector(self.activeLayerDidChange), with: nil, afterDelay: 0.0)
    }
    
    func selectedToolDidChange(_ note: Notification) {
        // the selected tool changed - find out which button cell matches and select it so that
        // the tool UI and the actual selected tool agree. This is necessary because when a tool is automatically
        // "sprung back" the UI needs to keep up with that automatic change.
        // which tool was selected?
        let tool: DKDrawingTool? = (note.object? as AnyObject).drawingTool()
        let toolName: String? = tool?.registeredName()
        // keep the "sticky" checkbox synchronised with the tool controller's actual state
        let sticky: Bool? = !(mDrawingView.controller as? DKToolController)?.automaticallyRevertsToSelectionTool()
        mToolStickyCheckbox = sticky
        // search through the matrix to find the cell whose title matches the tool's name,
        // and select it.
        var row: Int
        var col: Int
        var rr: Int
        var cc: Int
        var cell: NSCell?
        mToolMatrix.getNumberOfRows(row, columns: col)
        for rr in 0..<row {
            for cc in 0..<col {
                cell = mToolMatrix.cell(atRow: rr, column: cc)
                if (cell?.title == toolName) {
                    mToolMatrix.selectCell(atRow: rr, column: cc)
                    return
                }
            }
        }
        mToolMatrix.selectCell(atRow: 0, column: 0)
    }
    
    // MARK: -
    
    func updateControls(forSelection selection: [Any]) {
        // update all necessary UI controls to match the state of the selected object. Note that this ignores the selection passed to it
        // and just gets the info directly. It also doesn't bother to worry about more than one selected object - it just uses the info from
        // the topmost object - for this simple demo that's sufficient.
        // get the selected object's style
        let style: DKStyle? = styleOfSelectedObject()
        var rast: DKRasterizer?
        var temp: NSColor?
        var sw: Float
        // set up the fill controls if the style has a fill property, or disable them
        // altogether if it does not.
        if (style?.hasFill())! {
            rast = style?.renderersOfClass(DKFill.self)?.last as! DKRasterizer
            temp = (rast as? DKFill)?.colour()
            mStyleFillColourWell.isEnabled = true
            mStyleFillCheckbox = true
        }
        else {
            temp = NSColor.white
            mStyleFillColourWell.isEnabled = false
            mStyleFillCheckbox = false
        }
        mStyleFillColourWell.color = temp
        // set up the stroke controls if the style has a stroke property, or disable them
        // altogether if it does not.
        if (style?.hasStroke())! {
            rast = style?.renderersOfClass(DKStroke.self)?.last as! DKRasterizer
            temp = (rast as? DKStroke)?.colour()
            sw = (rast as? DKStroke)?.width
            mStyleStrokeColourWell.isEnabled = true
            mStyleStrokeWidthStepper.isEnabled = true
            mStyleStrokeWidthTextField.isEnabled = true
            mStyleStrokeCheckbox = true
        }
        else {
            temp = NSColor.white
            sw = 1.0
            mStyleStrokeColourWell.isEnabled = false
            mStyleStrokeWidthStepper.isEnabled = false
            mStyleStrokeWidthTextField.isEnabled = false
            mStyleStrokeCheckbox = false
        }
        mStyleStrokeColourWell.color = temp!
        mStyleStrokeWidthStepper = sw
        mStyleStrokeWidthTextField = sw
    }
    
    func styleOfSelectedObject() -> DKStyle {
        // returns the style of the topmost selected object in the active layer, or nil if there is nothing selected.
        var selectedStyle: DKStyle? = nil
        // get the active layer, but only if it's one that supports drawable objects
        let activeLayer = mDrawingView.drawing().activeLayer(of: DKObjectDrawingLayer.self)
        if activeLayer != nil {
            // get the selected objects and use the style of the last object, corresponding to the
            // one drawn last, or on top of all the others.
            let selectedObjects: [Any] = activeLayer.selectedAvailableObjects()
            if selectedObjects != nil && selectedObjects.count > 0 {
                selectedStyle = (selectedObjects.last as? DKDrawableObject)?.style
                selectedStyle?.locked = false
                // ensure it can be edited
            }
        }
        return selectedStyle!
    }
    
    // MARK: -
    
    // MARK: - as a NSWindowController
    
    override func awakeFromNib() {
        super.awakeFromNib()
        // make sure the view has a drawing object initialised. While the view itself would do this for us later, we tip its hand now so that we definitely
        // have a valid DKDrawing object available for setting up the notifications and user interface. In this case we are simply allowing the view to
        // create and own the drawing, rather than owning it here - though that would also be a perfectly valid way to do things.
        mDrawingView.createAutomaticDrawing()
        // subscribe to selection, layer and tool change notifications so that we can update the UI when these change
        NotificationCenter.default.addObserver(self, selector: #selector(self.drawingSelectionDidChange), name: NSNotification.Name(rawValue: kDKLayerSelectionDidChange), object: nil)
        NotificationCenter.default.addObserver(self, selector: #selector(self.drawingSelectionDidChange), name: NSNotification.Name.dkStyleDidChange, object: nil)
        NotificationCenter.default.addObserver(self, selector: #selector(self.activeLayerDidChange), name: NSNotification.Name(rawValue: kDKDrawingActiveLayerDidChange), object: mDrawingView.drawing())
        NotificationCenter.default.addObserver(self, selector: #selector(self.numberOfLayersChanged), name: NSNotification.Name(rawValue: kDKLayerGroupNumberOfLayersDidChange), object: mDrawingView.drawing())
        NotificationCenter.default.addObserver(self, selector: #selector(self.selectedToolDidChange), name: NSNotification.Name.dkDidChangeTool, object: nil)
        // creating the drawing set up the initial active layer but we weren't ready to listen to that notification. So that we can set
        // up the user-interface correctly this first time, just call the responder method directly now.
        activeLayerDidChange(nil)
        mDrawingView.window?.makeFirstResponder(mDrawingView)
    }
    
    // MARK: -
    
    // MARK: - as the TableView dataSource
    
    func numberOfRows(in aTable: NSTableView) -> Int {
        return mDrawingView.drawing().countOfLayers()
    }
    
    func tableView(_ aTableView: NSTableView, objectValueFor aTableColumn: NSTableColumn, row rowIndex: Int) -> Any {
        return mDrawingView.drawing().layers()[rowIndex].value(forKey: aTableColumn.identifier)
    }
    
    func tableView(_ aTableView: NSTableView, setObjectValue anObject: Any, for aTableColumn: NSTableColumn, row rowIndex: Int) {
        let layer: DKLayer? = (mDrawingView.drawing().layers()[rowIndex] as? DKLayer)
        layer?[aTableColumn.identifier] = anObject
    }
    
    // MARK: -
    
    // MARK: - as the TableView delegate
    
    func tableViewSelectionDidChange(_ aNotification: Notification) {
        // when the user selects a different layer in the table, change the real active layer to match.
        if aNotification.object == mLayerTable {
            let row: Int = mLayerTable.selectedRow
            if row != -1 {
                mDrawingView.drawing().activeLayer = mDrawingView.drawing().objectInLayers(at: row)
            }
        }
    }
    
    // MARK: -
    
    // MARK: - as the NSApplication delegate
    
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        // app ready to go - first turn off all style sharing. For this simple demo this makes life a bit easier.
        // (note - comment out this line and see what happens. It's perfectly safe ;-)
        DKStyle.stylesAreSharableByDefault = false
        // set up an initial style to apply to all new objects created. Because sharin gis off above, this style is copied
        // for each new object created, so each has its own individual style which can be edited independently.
        let ds = DKStyle(fillColour: NSColor.orange, strokeColour: NSColor.black, strokeWidth: 2.0)
        ds?.name = "Demo Style"
        DKObjectCreationTool.setStyleForCreatedObjects(ds)
        // register the default set of tools (Select, Rectangle, Oval, etc)
        DKDrawingTool.registerStandardTools()
    }
    
    @IBAction func saveDocument(as sender: Any) {
        let sp = NSSavePanel()
        sp.requiredFileType = "pdf"
        sp.canSelectHiddenExtension = true
        sp.beginSheet(forDirectory: nil, file: window?.title, modalForWindow: window, modalDelegate: self, didEndSelector: Selector("savePanelDidEnd:returnCode:contextInfo:"), contextInfo: nil)
    }
    
    func savePanelDidEnd(_ panel: NSSavePanel, returnCode: Int, contextInfo: UnsafeMutableRawPointer) {
        if returnCode == NSOKButton {
            let pdf: Data? = mDrawingView.drawing().pdf()
            pdf?.write(to: panel.url, options: true)
        }
    }
    
    // MARK: -
    
    // MARK: - as the Window delegate
    
    func windowWillReturnUndoManager(_ window: NSWindow) -> UndoManager {
        // DK's own implementation of the undo manager is generally more functional than the default Cocoa one, especially
        // for interactive drawing as it implements task coalescing.
        var um: DKUndoManager? = nil
        if um == nil {
            um = DKUndoManager()
        }
        return (um as? UndoManager)!
    }
}


